#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The main interface through which you should interact with the Strata API.
'''
import logging
import socket
from stratasdk.infrastructure import contextmanager
from stratasdk.infrastructure.connectionspool import ConnectionsPool
from stratasdk.infrastructure.proxy import Proxy
from stratasdk.infrastructure.contextmanager import Mode
from stratasdk.infrastructure.brokers import solutions
from stratasdk.infrastructure.brokers import articles
from stratasdk.infrastructure.brokers import cases
from stratasdk.infrastructure.brokers import comments
from stratasdk.infrastructure.brokers import attachments
from stratasdk.infrastructure.brokers import problems
from stratasdk.infrastructure.brokers import InstanceMaker

__author__ =  'Keith Robertson <kroberts@redhat.com>'

STREAM_LOG_FORMAT = '%(levelname)s: %(message)s'
USER_AGENT='strata-sdk-1.0.1'

class API():
    def __init__(self,
                 username,
                 password,
                 url='https://api.access.redhat.com',
                 key_file=None,
                 cert_file=None,
                 proxy_url=None,
                 proxy_user=None,
                 proxy_pass=None,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 httpDebug=False,
                 verbose=False,
                 userAgent=None):

        """
        To begin using the SDK you should instantiate an instance of the API.

        Keyword arguments:
        username   -- access.redhat.com user id
        password   -- password
        url        -- Strata REST URL (default: https://api.access.redhat.com)
        key_file   -- key_file for ssl enabled connection (not required)
        cert_file  -- cert_file for ssl enabled connection (not required)
        proxy_url  -- The url of the proxy server (e.g. http://example.com:8888/)
        proxy_user -- Optional username for HTTP basic auth to proxy server.
        proxy_pass -- Optional password for HTTP basic auth to proxy server.
        timeout    -- request timeout (not required)
        httpDebug  -- Print the HTTP request response flow. This is very useful if
                      you only want to see what you are sending and what you are receiving
                      from the REST API (default False)
        verbose    -- Print everything including HTTP request response flow.
        """

        # Set up logging

        if verbose:
            logging.basicConfig(level=logging.DEBUG, format=STREAM_LOG_FORMAT)
            httpDebug = verbose
        else:
            logging.basicConfig(level=logging.INFO, format=STREAM_LOG_FORMAT)

        if userAgent:
            ua = {'User-Agent': userAgent}
        else:
            ua =  {'User-Agent': USER_AGENT}
        contextmanager.add('proxy',
                           Proxy(ConnectionsPool(url=url,
                                                 key_file=key_file,
                                                 cert_file=cert_file,
                                                 timeout=timeout,
                                                 username=username,
                                                 password=password,
                                                 proxy_url=proxy_url,
                                                 proxy_user=proxy_user,
                                                 proxy_pass=proxy_pass,
                                                 debug=httpDebug),
                                 ua),
                           Mode.R)

        # Initialize the container classes.
        self.solutions = solutions(httpDebug)
        self.articles  = articles(httpDebug)
        self.cases     = cases(httpDebug)
        self.comments  = comments(httpDebug)
        self.attachments = attachments(httpDebug)
        self.problems  = problems(httpDebug)
        self.im        = InstanceMaker()



    def disconnect(self):
        ''' terminates server connection/s '''
        contextmanager._remove('proxy', force=True)


