#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Module that brokers the RESTful methods.
'''
import os.path
import logging
import socket
import mimetypes
from urlparse import urlparse
from xml.sax.saxutils import escape
from stratasdk.xml import params, report
from stratasdk.utils.urlhelper import UrlHelper
from stratasdk.utils.filterhelper import FilterHelper
from stratasdk.utils.parsehelper import ParseHelper, ReportParseHelper
from stratasdk.utils.searchhelper import SearchHelper
from stratasdk.infrastructure.common import Base
from stratasdk.infrastructure.errors import RequestError
from stratasdk.infrastructure.errors import ConnectionError

__author__ =  'Keith Robertson <kroberts@redhat.com>'

class solution(params.solution, Base):
    def __init__(self, solution):
        self.superclass = solution

    def __new__(cls, solution):
        if solution is None: return None
        obj = object.__new__(cls)
        obj.__init__(solution)
        return obj

    @classmethod
    def fromProps(cls, createdBy=None, createdDate=None, lastModifiedBy=None,
                  lastModifiedDate=None, id=None, uri=None, uidName=None,
                  view_uri=None, title=None, authorSSOName=None,
                  lastModifiedBySSOName=None, language=None, issue=None,
                  environment=None, resolution=None, rootCause=None,
                  internalDiagnosticSteps=None, externalDiagnosticSteps=None,
                  summary=None, tags=None, case=None, tag=None, published=None,
                  duplicateOf=None, body=None, kcsState=None,
                  ModerationState=None, explanation=None):

        if issue is not None:
            issue = params.issueType(issue, None)

        if resolution is not None:
            resolution = params.resolutionType(resolution, None)

        if environment is not None:
            environment = params.environmentType(environment, None)

        if rootCause is not None:
            rootCause = params.rootCauseType(rootCause, None)

        if internalDiagnosticSteps is not None:
            internalDiagnosticSteps = params.internalDiagnosticStepsType(internalDiagnosticSteps, None)

        if tag is not None:
            tag = params.tagType(tag, None)

        if duplicateOf is not None:
            duplicateOf = params.duplicateOfType(None, duplicateOf)

        psol = params.solution(createdBy, createdDate, lastModifiedBy,
                               lastModifiedDate, id, uri, uidName,
                               view_uri, title, authorSSOName,
                               lastModifiedBySSOName, language, issue,
                               environment, resolution, rootCause,
                               internalDiagnosticSteps, externalDiagnosticSteps,
                               summary, tags, case, tag, published,
                               duplicateOf, body, kcsState,
                               ModerationState, explanation)
        return cls(psol)

    def set_issue(self,issue):
        issue = params.issueType(issue, None)
        super(solution, self).set_issue(issue)

    def set_resolution(self, resolution):
        resolution = params.resolutionType(resolution, None)
        super(solution, self).set_resolution(resolution)

    def set_environment(self, environment):
        environment = params.environmentType(environment, None)
        super(solution, self).set_environment(environment)

    def set_rootCause(self, rootCause):
        rootCause = params.rootCauseType(rootCause, None)
        super(solution, self).set_rootCause(rootCause)

    def set_internalDiagnosticSteps(self,internalDiagnosticSteps):
        internalDiagnosticSteps = params.internalDiagnosticStepsType(internalDiagnosticSteps, None)
        super(solution, self).set_internalDiagnosticSteps(internalDiagnosticSteps)

    def set_tag(self, tag):
        tag = params.tagType(tag, None)
        super(solution, self).set_tag(tag)

    def set_duplicateOf(self, duplicateOf):
        duplicateOf = params.duplicateOfType(None, duplicateOf)
        super(solution, self).set_duplicateOf(duplicateOf)

    def update(self):
        '''
        Update this solution. This solution *must* have an ID.

        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/solutions'
        if self.get_id() is not None:
            self._getProxy().update(url=UrlHelper.append(url, self.superclass.get_id()),
                                    body=ParseHelper.toXml(self.superclass))
        else:
            raise RequestError('ID cannot be None on update')


class solutions(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def add(self, sol):
        '''
        Add a new solution

        Keyword arguments:
        sol -- The solution to be added.  Use InstanceMaker.makeSolution to get a solution bean
        return   -- The same solution with the ID and view_uri set if successful.
        raises   -- An exception if there was a connection related issue or an issue parsing
                    headers
        '''
        url = '/rs/solutions'
        doc, headers = self._getProxy().add(url=url,
                                            body=ParseHelper.toXml(sol))

        d = {}
        d.update(headers)
        view_uri = d['view-uri']
        parsed = urlparse(view_uri)
        sol.set_id(os.path.basename(parsed.path))
        sol.set_view_uri(view_uri)
        return sol


    def get(self, solutionID=None):
        '''
        Queries the API for the given solution ID.
        RESTful method: https://api.access.redhat.com/rs/solutions/<solutionID>

        Keyword arguments:
        solutionID -- The exact solutionID you are interested in.
        returns   -- A solution object that represents the given solution ID.  It has all
                    those attributes defined in the XSD.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/solutions'
        doc, headers = self._getProxy().get(url=UrlHelper.append(url, solutionID))
        sol = solution(doc)
        if self.httpDebug:
            print "REST Response:"
            sol.toXml()
        return sol

    def list(self, keywords=None, **kwargs):
        '''
        Queries the solutions RESTful interface with a given set of keywords.
        RESTful method: https://api.access.redhat.com/rs/solutions?keyword=NFS

        Keyword arguments:
        keywords -- A space separated list of keywords to search on.
        kwargs   -- Additional properties to filter on.  The RESTful interface can
                    only search on keywords; however, you can use this method to
                    post-filter the results returned.  Simply supply a list of
                    valid properties and their associated values.
                    Example: solutions.list('RHEV', authorSSOName="rhn-support-dconsoli")
        returns  -- A list of solution objects
        raises   -- An exception if there was a connection related issue.
        '''
        url='/rs/solutions'

        doc, headers = self._getProxy().get(url=SearchHelper.appendQuery(url, {'keyword':keywords}))
        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(solution,
                                        FilterHelper.filter(doc.get_solution(), kwargs))


class article(params.article, Base):
    def __init__(self, article):
        self.superclass = article

    def __new__(cls, article):
        if article is None: return None
        obj = object.__new__(cls)
        obj.__init__(article)
        return obj

    @classmethod
    def fromProps(cls, createdBy=None, createdDate=None, lastModifiedBy=None, lastModifiedDate=None,
                  id=None, uri=None, uidName=None, view_uri=None, title=None, authorSSOName=None,
                  lastModifiedBySSOName=None, language=None, issue=None, environment=None, resolution=None,
                  rootCause=None, internalDiagnosticSteps=None, articleBody=None, externalDiagnosticSteps=None,
                  summary=None, tags=None, case=None, tag=None, published=None, duplicateOf=None, kcsState=None,
                  body=None, explanation=None):

        if issue is not None:
            issue = params.issueType(issue, None)

        if resolution is not None:
            resolution = params.resolutionType(resolution, None)

        if environment is not None:
            environment = params.environmentType(environment, None)

        if rootCause is not None:
            rootCause = params.rootCauseType(rootCause, None)

        if internalDiagnosticSteps is not None:
            internalDiagnosticSteps = params.internalDiagnosticStepsType(internalDiagnosticSteps, None)

        if tag is not None:
            tag = params.tagType(tag, None)

        if duplicateOf is not None:
            duplicateOf = params.duplicateOfType(None, duplicateOf)

        psol = params.article(createdBy, createdDate, lastModifiedBy, lastModifiedDate,
                              id, uri, uidName, view_uri, title, authorSSOName,
                              lastModifiedBySSOName, language, issue, environment, resolution,
                              rootCause, internalDiagnosticSteps, articleBody, externalDiagnosticSteps,
                              summary, tags, case, tag, published, duplicateOf, kcsState,
                              body, explanation)
        return cls(psol)

    def set_issue(self,issue):
        issue = params.issueType(issue, None)
        super(article, self).set_issue(issue)

    def set_resolution(self, resolution):
        resolution = params.resolutionType(resolution, None)
        super(article, self).set_resolution(resolution)

    def set_environment(self, environment):
        environment = params.environmentType(environment, None)
        super(article, self).set_environment(environment)

    def set_rootCause(self, rootCause):
        rootCause = params.rootCauseType(rootCause, None)
        super(article, self).set_rootCause(rootCause)

    def set_internalDiagnosticSteps(self,internalDiagnosticSteps):
        internalDiagnosticSteps = params.internalDiagnosticStepsType(internalDiagnosticSteps, None)
        super(article, self).set_internalDiagnosticSteps(internalDiagnosticSteps)

    def set_tag(self, tag):
        tag = params.tagType(tag, None)
        super(article, self).set_tag(tag)

    def set_duplicateOf(self, duplicateOf):
        duplicateOf = params.duplicateOfType(None, duplicateOf)
        super(article, self).set_tag(duplicateOf)

## No such animal as delete.
#    def delete(self):
#        '''
#
#        '''
#        url = '/rs/articles'
#        sol = self._getProxy().delete(url=UrlHelper.append(url, self.get_id()))
#        print "----%s---" % sol


    def update(self):
        '''
        Update this article. This article *must* have an ID.

        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/articles'
        if self.get_id() is not None:
            self._getProxy().update(url=UrlHelper.append(url, self.superclass.get_id()),
                                    body=ParseHelper.toXml(self.superclass))
        else:
            raise RequestError('ID cannot be None on update')


class articles(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def add(self, art):
        '''
        Add a new article

        Keyword arguments:
        art -- The article to be added.  Use InstanceMaker.makeArticle to get a article bean
        return   -- The same article with the ID and view_uri set if successful.
        raises   -- An exception if there was a connection related issue or an issue parsing
                    headers
        '''
        url = '/rs/articles'
        doc, headers = self._getProxy().add(url=url,
                                      body=ParseHelper.toXml(art))
        d = {}
        d.update(headers)
        view_uri = d['view-uri']
        parsed = urlparse(view_uri)
        art.set_id(os.path.basename(parsed.path))
        art.set_view_uri(view_uri)
        return art

    def get(self, articleID=None):
        '''
        Queries the API for the given article ID.
        RESTful method: https://api.access.redhat.com/rs/articles/<articleID>

        Keyword arguments:
        articleID -- The exact articleID you are interested in.
        returns   -- A article object that represents the given article ID.  It has all
                    those attributes defined in the XSD.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/articles'

        doc, headers = self._getProxy().get(url=UrlHelper.append(url, articleID))
        art = article(doc)
        if self.httpDebug:
            print "REST Response:"
            art.toXml()
        return art

    def list(self, keywords=None, **kwargs):
        '''
        Queries the articles RESTful interface with a given set of keywords.
        RESTful method: https://api.access.redhat.com/rs/articles?keyword=NFS

        Keyword arguments:
        keywords -- A space separated list of keywords to search on.
        kwargs   -- Additional properties to filter on.  The RESTful interface can
                    only search on keywords; however, you can use this method to
                    post-filter the results returned.  Simply supply a list of
                    valid properties and their associated values.
                    Example: articles.list('RHEV', authorSSOName="rhn-support-dconsoli")
        returns  -- A list of article objects
        raises   -- An exception if there was a connection related issue.
        '''
        url='/rs/articles'

        doc, headers = self._getProxy().get(url=SearchHelper.appendQuery(url, {'keyword':keywords}))
        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(article,
                                        FilterHelper.filter(doc.get_article(), kwargs))

class comment(params.comment, Base):
    def __init__(self, comment):
        self.superclass = comment

    def __new__(cls, comment):
        if comment is None: return None
        obj = object.__new__(cls)
        obj.__init__(comment)
        return obj

    @classmethod
    def fromProps(cls, createdBy=None, createdDate=None, lastModifiedBy=None,
                  lastModifiedDate=None, id=None, caseNumber=None, text=None,
                  uri=None, public=True, draft=False, publishedDate=None):

        comment = params.comment(createdBy, createdDate, lastModifiedBy,
                                 lastModifiedDate, id, caseNumber, text,
                                 uri, public, draft, publishedDate)
        return cls(comment)


    def update(self):
        '''
        Update this comment. This comment must have both an comment ID and a case ID.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/comments/{commentID}'


        if self.get_id() is not None and self.get_id() is not None:
            url = UrlHelper.replace(url,
                                    {'{caseNumber}': self.get_caseNumber(),
                                     '{commentID}': self.get_id()})
            self._getProxy().update(url,
                                    body=ParseHelper.toXml(self))
        else:
            raise RequestError('ID cannot be None on update')



class comments(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def add(self, cmnt):
        '''
        Add a new comment

        Keyword arguments:
        cmnt     -- The comment to be added.  Use InstanceMaker.makeComment to get a bean.
        return   -- The same comment with the ID and view_uri set if successful.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/comments'

        url = UrlHelper.replace(url,{'{caseNumber}': cmnt.get_caseNumber()})

        doc, headers = self._getProxy().add(url=url,
                                            body=ParseHelper.toXml(cmnt))
        d = {}
        d.update(headers)
        view_uri = d['view_uri']
        parsed = urlparse(view_uri)
        cmnt.set_caseNumber(os.path.basename(parsed.path))
        return cmnt

    def get(self, caseNumber=None, commentID=None):
        '''
        Queries the API for the given solution ID.
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>/comments/<commentID>

        Keyword arguments:
        caseNumber -- The exact caseNumber you are interested in.
        commentID  -- The exact comment ID you're interested in (e.g.a0aA00000079GpQIAU)
        returns   -- A comment object that represents the given comment ID.  It has all
                    those attributes defined in the XSD.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/comments/{commentID}'

        url = UrlHelper.replace(url,
                                {'{caseNumber}': caseNumber,
                                 '{commentID}': commentID})

        doc, headers = self._getProxy().get(url)
        com = comment(doc)
        if self.httpDebug:
            print "REST Response:"
            com.toXml()
        return com

    def list(self, caseNumber=None, startDate=None, endDate=None, **kwargs):
        '''
        Gets all of the comments for a given case number.  You can then
        search/filter the returned comments using any of the properties
        of a 'comment'
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>/comments.

        Keyword arguments:
        caseNumber -- A case number (e.g.00595293)
        startDate --  Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        endDate   -- Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        kwargs   -- Additional properties to filter on.  The RESTful interface can
                    only search on keywords; however, you can use this method to
                    post-filter the results returned.  Simply supply a list of
                    valid properties and their associated values.
                    Example: comments.list('00595293', createdBy="rhn-support-dconsoli")
        returns  -- A list of comment objects
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/comments'

        url = UrlHelper.replace(url,
                                {'{caseNumber}': caseNumber})

        doc, headers = self._getProxy().get(SearchHelper.appendQuery(url,
                                                                     {'startDate':startDate,
                                                                      'endDate':endDate}))
        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(comment,
                                        FilterHelper.filter(doc.get_comment(), kwargs))

class attachment(params.attachment, Base):
    def __init__(self, attachment):
        self.superclass = attachment

    def __new__(cls, attachment):
        if attachment is None: return None
        obj = object.__new__(cls)
        obj.__init__(attachment)
        return obj

    @classmethod
    def fromProps(cls, createdBy=None, createdDate=None, lastModifiedBy=None,
                  lastModifiedDate=None, caseNumber=None, caseId=None,
                  uuid=None, etag=None, uri=None, fileDate=None, fileName=None,
                  description=None, mimeType=None, length=None, active=None,
                  deprecated=None, private=False):

        attachment = params.attachment(createdBy, createdDate, lastModifiedBy,
                                       lastModifiedDate, caseNumber, caseId,
                                       uuid, etag, uri, fileDate, fileName,
                                       description, mimeType, length, active,
                                       deprecated, private)
        return cls(attachment)





class attachments(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def add(self, caseNumber=None, public=True, fileName=None, description=None):
        '''
        Add a new comment

        Keyword arguments:
        caseNumber -- The exact caseNumber you are interested in.
        public     -- A public or private attachment.
        fileName   -- Either a file in the CWD or a full path to a file.
        return     -- The same comment with the ID and view_uri set if successful.
        raises     -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/attachments'
        fh = None

        url = UrlHelper.replace(url,{'{caseNumber}': caseNumber})

        doc, headers = self._getProxy().upload(SearchHelper.appendQuery(url, {'public':public}),
                                               fileName,
                                               description)
        amnt = attachment(doc)
        amnt.set_private(not public)
        amnt.set_mimeType(mimetypes.guess_type(fileName)[0] or 'application/octet-stream')
        if self.httpDebug:
            print "REST Response:"
            amnt.toXml()
        return amnt


    def delete(self, caseNumber=None, attachmentUUID=None):
        '''
        Removes the attachment from the case.
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>/attachments/<attachmentUUID>

        Keyword arguments:
        caseNumber      -- The exact caseNumber you are interested in.
        attachmentUUID  -- The exact attachment UUID you're interested in.
        returns         -- true if successful
        raises          -- An exception if the attachment was not able to be removed.
        '''
        url = '/rs/cases/{caseNumber}/attachments/{attachmentUUID}'

        url = UrlHelper.replace(url,
                                {'{caseNumber}': caseNumber,
                                 '{attachmentUUID}': attachmentUUID})

        doc, headers = self._getProxy().delete(url)
        return True


    def get(self, caseNumber=None, attachmentUUID=None, fileName=None, destDir=None):
        '''
        Queries the API for the given attachment.
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>/attachments/<attachmentUUID>

        Keyword arguments:
        caseNumber      -- The exact caseNumber you are interested in.
        attachmentUUID  -- The exact attachment UUID you're interested in.
        fileName        -- The name you want to give the file after it has been downloaded.
        destDir         -- The directory which the attachment should be saved in.
        returns         -- The path to the saved file as a string.
        raises          -- An exception if there was a connection, file open, etc. issue.
        '''
        url = '/rs/cases/{caseNumber}/attachments/{attachmentUUID}'

        # Pre-checks for proper input
        if not caseNumber or not attachmentUUID or not fileName:
            raise Exception('caseNumber(%s) attachmentUUID(%s) fileName(%s) cannot be empty.' %
                            (caseNumber, attachmentUUID, fileName))

        # Join the path
        if destDir:
            fileName = os.path.join(destDir, fileName)

        url = UrlHelper.replace(url,
                                {'{caseNumber}': caseNumber,
                                 '{attachmentUUID}': attachmentUUID})
        conn = None
        fh = None
        try:
            logging.debug('Downloading %s...' % fileName)
            conn = self._getProxy().getConnectionsPool().getConnection()
            conn.doRequest(method='GET', url=url)
            response = conn.getResponse()
            if response.status < 400:
                fh = open(fileName, 'wb')
                fh.write(response.read())
                logging.debug('Successfully downloaded %s.' % fileName)
            else:
                logging.debug("ERROR: unable to download %s as %s. Reason: %s" %
                              (url, fileName, response.reason))
                raise RequestError(response.status, response.reason, None)
        except socket.error, se:
            logging.debug('Socket error: msg(%s)' % (se))
            raise ConnectionError, str(se)
        except IOError, ioe:
            logging.debug('I/O error: errno(%s) strerror(%s)' % (ioe.errno, ioe.strerror))
            raise ioe
        except Exception, e:
            logging.debug('Unexpected exception: msg(%s)' % (str(e)))
            raise e
        finally:
            if conn:
                try:
                    conn.close()
                except Exception:
                    pass
            if fh:
                try:
                    fh.close()
                except Exception:
                    pass

        return fileName


    def list(self, caseNumber=None, startDate=None, endDate=None, **kwargs):
        '''
        Gets all of the attachments for a given case number.  You can then
        search/filter the returned comments using any of the properties
        of a 'attachment'
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>/attachments.

        Keyword arguments:
        caseNumber -- A case number (e.g.00595293)
        startDate --  Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        endDate   -- Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        kwargs   -- Additional properties to filter on.  The RESTful interface can
                    only search on keywords; however, you can use this method to
                    post-filter the results returned.  Simply supply a list of
                    valid properties and their associated values.
                    Example: comments.list('00595293', createdBy="rhn-support-dconsoli")
        returns  -- A list of comment objects
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases/{caseNumber}/attachments'

        url = UrlHelper.replace(url,
                                {'{caseNumber}': caseNumber})

        doc, headers = self._getProxy().get(SearchHelper.appendQuery(url,
                                                                     {'startDate':startDate,
                                                                      'endDate':endDate}))
        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(attachment,
                                        FilterHelper.filter(doc.get_attachment(), kwargs))


class entitlement(params.entitlement, Base):
    def __init__(self, entitlement):
        self.superclass = entitlement

    def __new__(cls, entitlement):
        if entitlement is None: return None
        obj = object.__new__(cls)
        obj.__init__(entitlement)
        return obj

    @classmethod
    def fromProps(cls, createdBy=None, createdDate=None, lastModifiedBy=None,
                  lastModifiedDate=None, id=None, uri=None, name=None, sla=None,
                  supportLevel=None, serviceLevel=None, startDate=None, endDate=None):

        entitlement = params.entitlement(createdBy, createdDate, lastModifiedBy,
                                         lastModifiedDate, id, uri, name,
                                         sla, supportLevel, serviceLevel,
                                         startDate, endDate)
        return cls(entitlement)

class case(params.case, Base):
    def __init__(self, case):
        self.superclass = case

    def __new__(cls, case):
        if case is None: return None
        obj = object.__new__(cls)
        obj.__init__(case)
        return obj

    @classmethod
    def fromProps(cls, summary=None, product=None, version=None, **kwargs):
        case = params.case(summary=summary, product=product, version=version, **kwargs)
        return cls(case)

    def update(self):
        '''
        Update this case. This case *must* have an case number.

        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases'
        if self.get_caseNumber() is not None:
            self._getProxy().update(url=UrlHelper.append(url, self.get_caseNumber()),
                                    body=ParseHelper.toXml(self.superclass))
        else:
            raise RequestError('ID cannot be None on update')

    def get_comments(self):
        return ParseHelper.toCollection(comment,
                                        params.case.get_comments(self).comment)
    def get_solutions(self):
        return ParseHelper.toCollection(solution,
                                        params.case.get_solutions(self).solution)
    def get_entitlement(self):
        return entitlement(params.case.get_entitlement(self))



class cases(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def add(self, cs):
        '''
        Add a new solution

        Keyword arguments:
        cs       -- The case to be added.  Use InstanceMaker.makeCase to get a case bean.
        return   -- The same case with the case number and view_uri set if successful.
        raises   -- An exception if there was a connection related issue.
        '''

        url = '/rs/cases'
        doc, headers = self._getProxy().add(url=url,
                                            body=ParseHelper.toXml(cs))
        d = {}
        d.update(headers)
        view_uri = d['view_uri']
        parsed = urlparse(view_uri)
        cs.set_caseNumber(os.path.basename(parsed.path))
        cs.set_view_uri(view_uri)
        return cs

    def get(self, caseNumber=None):
        '''
        Queries the API for the given case number.
        RESTful method: https://api.access.redhat.com/rs/cases/<caseNumber>

        Keyword arguments:
        caseNumber -- The exact caseNumber you are interested in.
        returns   -- A case object that represents the given case number.  It has all
                    those attributes defined in the XSD.
        raises   -- An exception if there was a connection related issue.
        '''
        url = '/rs/cases'

        doc, headers = self._getProxy().get(url=UrlHelper.append(url, caseNumber))
        cs = case(doc)
        if self.httpDebug:
            print "REST Response:"
            cs.toXml()
        return cs

    def list(self, keywords=None, includeClosed=False, detail=False,
             group=None, startDate=None, endDate=None, **kwargs):
        '''
        Queries the cases RESTful interface with a given set of keywords.
        RESTful method: https://api.access.redhat.com/rs/cases?keyword=NFS

        Keyword arguments:
        keywords -- A space separated list of keywords to search on.
        includeClosed -- Do not include closed cases.
        detail   -- Include additional details.
        group    -- See https://api.access.redhat.com/rs/groups
        startDate --  Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        endDate   -- Must be either: yyyy-MM-ddTHH:mm:ss or yyyy-MM-dd
        kwargs   -- Additional properties to filter on.  The RESTful interface can
                    only search on keywords; however, you can use this method to
                    post-filter the results returned.  Simply supply a list of
                    valid properties and their associated values.
                    Example: cases.list('RHEV', authorSSOName="rhn-support-dconsoli")
        returns  -- A list of solution objects
        raises   -- An exception if there was a connection related issue.
        '''
        url='/rs/cases'

        doc, headers = self._getProxy().get(url=SearchHelper.appendQuery(url,
                                                                   {'query':keywords,
                                                                    'includeClosed':includeClosed,
                                                                    'detail':detail,
                                                                    'group':group,
                                                                    'startDate':startDate,
                                                                    'endDate':endDate}))
        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(case,
                                        FilterHelper.filter(doc.get_case(), kwargs))


class problem(params.problem, Base):
    def __init__(self, problem):
        self.superclass = problem

    def __new__(cls, problem):
        if problem is None: return None
        obj = object.__new__(cls)
        obj.__init__(problem)
        return obj

    @classmethod
    def fromProps(cls, source=None, link=None):

        problem = params.problem(source, link)
        return cls(problem)

class problems(Base):
    def __init__(self, httpDebug=False):
        """Constructor."""
        # Assign self to superclass for the broker classes that are aggregates of the specific type.
        # Without this assignment you'll get a nice and confusing recursion stacktrace if
        # a wayward programmer attempts to call a non-existent method on this class.
        self.superclass = self
        self.httpDebug = httpDebug

    def diagnoseStr(self,
                    content=None):
        '''
        Queries the problems RESTful interface with a given string.
        RESTful method: https://api.access.redhat.com/rs/problems

        Keyword arguments:
        content -- A string whose content you want diagnosed.
        returns -- An array of problems.
        '''
        url='/rs/problems'


        rpt = report.report()

        rpt.add_binding(report.binding(type_='text',
                                       name='summary',
                                       valueOf_=escape(content)))
        doc, headers = self._getProxy().action(url,
                                               body=ReportParseHelper.toXml(rpt))

        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(problem, doc.get_problem())


    def diagnoseFile(self, fileName=None):
        '''
        Queries the problems RESTful interface with a given file.
        RESTful method: https://api.access.redhat.com/rs/problems

        Keyword arguments:
        content -- This can either be a file handle or a string whose content you want diagnosed.
        '''
        url='/rs/problems'
        fh = None
        try:
            fh = open(fileName, 'rb')
            doc, headers = self._getProxy().action(url,
                                               fh)
        except socket.error, se:
            logging.debug('Socket error: msg(%s)' % (se))
            raise ConnectionError, str(se)
        except IOError, ioe:
            logging.debug('I/O error: errno(%s) strerror(%s)' % (ioe.errno, ioe.strerror))
            raise ioe
        except Exception, e:
            logging.debug('Unexpected exception: msg(%s)' % (str(e)))
            raise e
        finally:
            if fh:
                fh.close()

        if self.httpDebug:
            print "REST Response:\n%s" % ParseHelper.toXml(doc)
        return ParseHelper.toCollection(problem, doc.get_problem())

class InstanceMaker():
    '''
    Utility class to make single instances of case, solution, article, entitlement, and comment.
    This is useful so that you only need import API and not all of the helper classes used by
    API to get the job done.
    '''
    @classmethod
    def makeCase(self, summary=None, product=None, version=None, **kwargs):
        return case.fromProps(summary=summary, product=product, version=version, **kwargs)
    @classmethod
    def makeSolution(self,createdBy=None, createdDate=None, lastModifiedBy=None,
                     lastModifiedDate=None, id=None, uri=None, uidName=None,
                     view_uri=None, title=None, authorSSOName=None,
                     lastModifiedBySSOName=None, language=None, issue=None,
                     environment=None, resolution=None, rootCause=None,
                     internalDiagnosticSteps=None, externalDiagnosticSteps=None,
                     summary=None, tags=None, case=None, tag=None, published=None,
                     duplicateOf=None, body=None, kcsState=None,
                     ModerationState=None, explanation=None):
        return solution.fromProps(createdBy, createdDate, lastModifiedBy,
                                  lastModifiedDate, id, uri, uidName,
                                  view_uri, title, authorSSOName,
                                  lastModifiedBySSOName, language, issue,
                                  environment, resolution, rootCause,
                                  internalDiagnosticSteps, externalDiagnosticSteps,
                                  summary, tags, case, tag, published,
                                  duplicateOf, body, kcsState,
                                  ModerationState, explanation)
    @classmethod
    def makeArticle(self,createdBy=None, createdDate=None, lastModifiedBy=None, lastModifiedDate=None,
                    id=None, uri=None, uidName=None, view_uri=None, title=None, authorSSOName=None,
                    lastModifiedBySSOName=None, language=None, issue=None, environment=None, resolution=None,
                    rootCause=None, internalDiagnosticSteps=None, articleBody=None, externalDiagnosticSteps=None,
                    summary=None, tags=None, case=None, tag=None, published=None, duplicateOf=None, kcsState=None,
                    body=None, explanation=None):
        return article.fromProps(createdBy, createdDate, lastModifiedBy, lastModifiedDate,
                                 id, uri, uidName, view_uri, title, authorSSOName,
                                 lastModifiedBySSOName, language, issue, environment, resolution,
                                 rootCause, internalDiagnosticSteps, articleBody, externalDiagnosticSteps,
                                 summary, tags, case, tag, published, duplicateOf, kcsState,
                                 body, explanation)
    @classmethod
    def makeEntitlement(self,createdBy=None, createdDate=None, lastModifiedBy=None,
                        lastModifiedDate=None, id=None, uri=None, name=None, sla=None,
                        supportLevel=None, serviceLevel=None, startDate=None, endDate=None):
        return entitlement.fromProps(createdBy, createdDate, lastModifiedBy,
                                     lastModifiedDate, id, uri, name, sla,
                                     supportLevel, serviceLevel, startDate, endDate)
    @classmethod
    def makeComment(self, createdBy=None, createdDate=None, lastModifiedBy=None,
                    lastModifiedDate=None, id=None, caseNumber=None, text=None,
                    uri=None, public=True, draft=False, publishedDate=None):
        return comment.fromProps(createdBy, createdDate, lastModifiedBy, lastModifiedDate, id, caseNumber,
                                 text, uri, public, draft, publishedDate)

    @classmethod
    def makeAttachment(self, createdBy=None, createdDate=None, lastModifiedBy=None,
                       lastModifiedDate=None, caseNumber=None, caseId=None,
                       uuid=None, etag=None, uri=None, fileDate=None, fileName=None,
                       description=None, mimeType=None, length=None, active=None,
                       deprecated=None, private=False):
        return comment.fromProps(createdBy, createdDate, lastModifiedBy,
                                 lastModifiedDate, caseNumber, caseId,
                                 uuid, etag, uri, fileDate, fileName,
                                 description, mimeType, length, active,
                                 deprecated, private)
