#
# Copyright (c) 2010 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#           http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from httplib import HTTPConnection, HTTPSConnection
import base64
import mimetypes
import os
import socket
import time
import urllib
import urlparse

class Connection(object):
    '''
    The strata api connection proxy
    '''
    def __init__(self,
                 url,
                 username,
                 password,
                 manager,
                 key_file=None,
                 cert_file=None,
                 strict=None,
                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 proxy_url=None,
                 proxy_user=None,
                 proxy_pass=None,
                 debug=False):

        self.__connection = self.__createConnection(url=url,
                                                    username=username,
                                                    password=password,
                                                    proxy_url=proxy_url,
                                                    proxy_user=proxy_user,
                                                    proxy_pass=proxy_pass,
                                                    key_file=key_file,
                                                    cert_file=cert_file,
                                                    strict=strict,
                                                    timeout=timeout)
        self.__connection.set_debuglevel(int(debug))
        self.__manager = manager
        self.__id = id(self)


    def get_id(self):
        return self.__id

    def getConnection(self):
        return self.__connection

    def getDefaultHeaders(self):
        return self.__headers.copy()

    def doRequest(self, method, url, body=urllib.urlencode({}), headers={}):
        return self.__connection.request(method, url, body, self.getHeaders(headers))

    def doUpload(self, url, fileName, description=None):
        '''
        Do an upload of a single file to the given URL.
        Keyword arguments:
            url -- The URL to upload to.
            fileHandle -- An open file handle whose content needs to be sent.
            description -- Optional description of the file.
        '''
        boundary = '----------%x' % time.time()
        fh = None
        try:
            fh = open(fileName, 'rb')
            # Compose the head of the form.
            headFormAry = []
            if description:
                headFormAry.append('--' + boundary)
                headFormAry.append('Content-Disposition: form-data; name="description"')
                headFormAry.append('')
                headFormAry.append(description)
            headFormAry.append('--' + boundary)
            headFormAry.append('Content-Disposition: form-data; name="%s"; filename="%s"' % ('file', fileName))
            headFormAry.append('Content-Type: %s' % (mimetypes.guess_type(fileName)[0] or 'application/octet-stream'))
            headFormAry.append('')
            headFormAry.append('')
            headForm = '\r\n'.join(headFormAry)

            # Compose the tail
            tailFormAry = []
            tailFormAry.append('')
            tailFormAry.append('--' + boundary + '--')
            tailFormAry.append('')
            tailForm = '\r\n'.join(tailFormAry)


            totalLength = len(headForm) + os.fstat(fh.fileno()).st_size + len(tailForm)

            # Start sending.
            self.__connection.putrequest(method='POST', url=url, skip_host=0, skip_accept_encoding=1)
            hdrDict = self.getHeaders({'Content-Length' : str(totalLength),
                                       'Content-Type' : 'multipart/form-data; boundary=%s' % boundary,
                                       'Accept' : 'text/plain'})
            for key, value in hdrDict.items():
                self.__connection.putheader(key , value)
            self.__connection.endheaders()
            self.__connection.send(headForm)
            self.__connection.send(fh)
            self.__connection.send(tailForm)
        except Exception, e:
            raise e
        finally:
            if fh:
                fh.close()


    def getHeaders(self, headers):
        extended_headers = self.getDefaultHeaders()
        for k in headers.keys():
            if (headers[k] is None and extended_headers.has_key(k)):
                extended_headers.pop(k)
            else:
                extended_headers[k] = headers[k]
        return extended_headers

    def getResponse(self):
        return self.__connection.getresponse()

    def setDebugLevel(self, level):
        self.__connection.set_debuglevel(level)

    def setTunnel(self, host, port=None, headers=None):
        self.__connection.set_tunnel(host, port, headers)

    def close(self):
        self.__connection.close()
#FIXME: create connection watchdog to close it on idle-ttl expiration, rather than after the call
        if (self.__manager is not None):
            self.__manager._freeResource(self)

    def state(self):
        return self.__connection.__state


    def __parse_url(self, url):
        if not url.startswith('http'):
            url = "https://" + url
        return urlparse.urlparse(url)

    def __createConnection(self,
                           url,
                           username,
                           password,
                           proxy_url=None,
                           proxy_user=None,
                           proxy_pass=None,
                           key_file=None,
                           cert_file=None,
                           strict=None,
                           timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        '''
        Create  a proxied or unproxied HTTP or HTTPS connection.  This will
        also set the default HTTP headers for the target.
        '''
        def makeConnection(scheme,
                           hostname,
                           port,
                           key_file=None,
                           cert_file=None,
                           strict=None,
                           timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
            if(scheme == 'https'):
                return HTTPSConnection(host=hostname,
                                       port=port,
                                       key_file=key_file,
                                       cert_file=cert_file,
                                       strict=strict,
                                       timeout=timeout)
            return HTTPConnection(host=hostname,
                                  port=port,
                                  strict=strict,
                                  timeout=timeout)


        conn = None
        u = self.__parse_url(url)

        if proxy_url:
            purl = self.__parse_url(proxy_url)
            conn = makeConnection(u.scheme, # Scheme must == dest url's scheme.
                                  purl.hostname,
                                  purl.port,
                                  key_file,
                                  cert_file,
                                  strict,
                                  timeout)
            # Set tunnel so httplib will send 'CONNECT example.com HTTP/1.0\r\n'
            hdr = {'Host': u.hostname,
                   'Proxy-Connection': 'Keep-Alive',
                   'Accept' : 'application/xml'}
            if proxy_user and proxy_pass:
                auth = base64.encodestring("%s:%s" % (proxy_user, proxy_pass)).strip()
                hdr['Proxy-authorization'] = "Basic %s" % auth
            # Critical step.  Proxy must know where to go.
            conn.set_tunnel(u.hostname,
                            None,
                            hdr)
        else:
            # Not proxied
            conn = makeConnection(u.scheme,
                                  u.hostname,
                                  u.port,
                                  key_file,
                                  cert_file,
                                  strict,
                                  timeout)

        # Create the default set of headers
        self.__headers = self.__createHeaders(username,
                                              password,
                                              u.hostname)
        return conn



    def __createHeaders(self, username, password, host):
        auth = base64.encodestring("%s:%s" % (username, password)).strip()
        return {"Accept": "application/xml",
                "Content-Type": "application/xml",
                'Host' : host,
                "Authorization": "Basic %s" % auth}

    id = property(get_id, None, None, None)
